/* ***********************************************
 
	GEMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemv_n_kernel.cu"
#include "mublas_gemv_t_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *x,
	const size_t incx,
	const TYPE0 *beta,
	TYPE0 *y,
	const size_t incy
) {
	uint32_t kernel = 0;
	uint32_t use_vector = 0;
	dim3 threads;
	dim3 grid;
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	if (*alpha == ZERO0) {
		MUBLAS_SCAL (mh, m, beta, y, incy);
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)x%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)y%(128/sizeof(TYPE0)))%16 == 0) &&
		(lda%mh->N_T == 0)
		) {
		use_vector = 1;
	} else {
		use_vector = 0;
	}
	mh->m = m;
	mh->n = n;
	// ---------------------------------------------

	// ---------------------------------------------
	kernel = 999; // default=CUBLAS;
	if (incx == 1 && incy == 1) {
		if (trans == 'N' || trans == 'n') {
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_N;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_N;
			mh->func_getTxMax     = &getTxMax_GEMV_N;
			#include "mublas_gemv_n.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 140;
				else			kernel = 141;
			} else if (mh->N_T == 2) { 
				if (use_vector) kernel = 120;
				else 			kernel = 121;
			} else {
								kernel = 110;
			}
		} else if (trans == 'T' || trans == 't') {
			mh->BlkDim = 1;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_T;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_T;
			mh->func_getTxMax     = &getTxMax_GEMV_T;
			#include "mublas_gemv_t.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 240;
				else			kernel = 210;
			} else if (mh->N_T == 2) {
				if (use_vector) kernel = 220;
				else			kernel = 210;
			} else {
								kernel = 210;
			}
		}
	}
	// ---------------------------------------------

	// ---------------------------------------------
	switch (kernel) {
		// N-kernel
		case 140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v4, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_N_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 141:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s4, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_N_DEVICE_s4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_v2, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_N_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 121:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s2, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_N_DEVICE_s2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_N_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_N_DEVICE_s1, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_N_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		// T-kernel
		case 240:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v4, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 220:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_v2, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		case 210:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_T_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_T_DEVICE_s1, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			else MUBLAS_GEMV_T_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y);
			break;
		// CUBLAS
		default:
			#if defined (CUBLAS_GEMV)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetStream (ch, mh->streamId);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasOperation_t tranAcublas = CUBLAS_OP_N;
			if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
			if (trans == 'C' || trans == 'c') tranAcublas = CUBLAS_OP_C;
			CUBLAS_GEMV (ch, tranAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------

	mh->kernelnumber = kernel;

	return 0;
}

__host__ int32_t
MUBLAS_GEMV_MGPU (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 alpha,
	TYPE0 **a,
	const size_t lda,
	TYPE0 **x,
	const size_t incx,
	const TYPE0 beta,
	TYPE0 **y,
	const size_t incy,
	const uint32_t ngpus,
	const size_t bs
) {
	uint32_t igpu;
	size_t nr, ni[NGPU_MAX] = {0};
	TYPE0 betai;

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	if (trans == 'n' || trans == 'N') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			betai = ZERO0;
			if (ni[igpu] == 0) {
				MUBLAS_SCAL (&mh[igpu], m, &betai, y[igpu], 1);
			} else {
				if (igpu == 0) betai = beta;
				MUBLAS_GEMV (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
			}
		}
	} else {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = ZERO0;
			if (igpu == 0) betai = beta;
			MUBLAS_SCAL (&mh[igpu], n, &betai, y[igpu], 1);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			if (igpu == 0) betai = ONE0;
			if (ni[igpu] != 0) MUBLAS_GEMV (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
		}
	}

	for(igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize();
	}
	return 0;
}

__host__ int32_t
MUBLAS_GEMV_MGPU_ASYNC (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 alpha,
	TYPE0 **a,
	const size_t lda,
	TYPE0 **x,
	const size_t incx,
	const TYPE0 beta,
	TYPE0 **y,
	const size_t incy,
	const uint32_t ngpus,
	const size_t bs
) {
	uint32_t igpu;
	size_t nr, ni[NGPU_MAX] = {0};
	TYPE0 betai;

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	if (trans == 'n' || trans == 'N') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			betai = ZERO0;
			if (ni[igpu] == 0) {
				MUBLAS_SCAL (&mh[igpu], m, &betai, y[igpu], 1);
			} else {
				if (igpu == 0) betai = beta;
				MUBLAS_GEMV (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
			}
		}
	} else {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = beta;
			//if (igpu == 0) betai = beta;
			//MUBLAS_SCAL (&mh[igpu], n, &betai, y[igpu], 1);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			//if (igpu == 0) betai = ONE0;
			if (ni[igpu] != 0) MUBLAS_GEMV (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
		}
	}
	return 0;
}


// note: Timer
/*
	mh->flops_prev = mh->flops;
	mh->WrpOcp_prev = mh->WrpOcp;
	mh->BlkOcp_prev = mh->BlkOcp;

	float exec_time;
	cudaEvent_t t_start, t_stop;
	cudaEventCreate (&t_start);
	cudaEventCreate (&t_stop);
	cudaEventRecord (t_start, 0);

// hoge ();

	cudaEventRecord (t_stop, 0);
	cudaEventSynchronize (t_stop);
	cudaEventElapsedTime (&exec_time, t_start, t_stop);
	mh->flops = 1.0e-6 * (2.*(float)m*(float)n+2.*(float)m+2.*(float)m) / exec_time;

	if (mh->GrdOcp > 0.95) {
		if (mh->flops < mh->flops_prev) {
		// performance down
			if (mh->WrpOcp_prev > mh->WrpOcp) 
				mh->WrpOcpMin = mh->WrpOcp_prev;
			if (mh->BlkOcp_prev > mh->BlkOcp) 
				mh->BlkOcpMin = mh->BlkOcp_prev;
		} else {
		// performance up
			if (mh->WrpOcp_prev > mh->WrpOcp) 
				mh->WrpOcpMin = mh->WrpOcp;
			if (mh->BlkOcp_prev > mh->BlkOcp) 
				mh->BlkOcpMin = mh->BlkOcp;
		}
	}
*/

