/* ***********************************************
 
	GEMV-N for CUDA 

   *********************************************** */   

#include "mublas_gemv_n.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_GEMV_N
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_X_GEMV_N
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif
#define LOAD1_Y LOAD1___
#define LOAD2_Y LOAD2___
#define LOAD4_Y LOAD4___
#define STORE1_Y STORE1___
#define STORE2_Y STORE2___
#define STORE4_Y STORE4___


/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_GEMV_N ( mublasHandle_t* mh) {
	mh->Bx = ceilingEx ((float)mh->m / (float)(mh->N_T * mh->Tx), 1);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_GEMV_N ( mublasHandle_t* mh) {
	mh->ShmPerBlk = ceilingEx ((float)(mh->Tx * mh->Ty * mh->N_T * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_GEMV_N ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->m / (float)mh->N_T / (float)mh->devprop_multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	kernel=s1, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_s1 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx);
	SIZE_T i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE1 Ar1, Xr1, Yr1, Tr1;

	Tr1 = ZERO1;

	// main
	for (i = 0; i < n/Th; i++) {
		k = i*Th;
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < Th; j+= Ty) 
				Tr1 = MAD1 (Xs1[j], Ag[(j+k) * lda + rowid], Tr1);
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th;
	if (k + iTh < n) {
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty) {
			if (j < n) {
				LOAD1_X (Ar1, &Ag[j * lda + rowid]);
				Tr1 = MAD1 (Xs1[j-k], Ar1, Tr1);
			}
		}
	}
	__syncthreads ();

	// reduction
	Xs1[iTh] = Tr1;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) 
			Tr1 = ADD1 (Tr1, Xs1[Tx * i + iTx]);
		if (rowid < m) {
			LOAD1_Y (Yr1, &Yg[rowid]);
			Yr1 = MAD1 (alpha, Tr1, MUL1 (beta, Yr1));
			STORE1_Y (&Yg[rowid], Yr1);
		}
	}
#endif
}

/* ***********************************************
	kernel=v2, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_v2 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k;
	extern __shared__ TYPE2 Xs2[];
	TYPE2 Ar2, Xr2, Yr2, Tr2;
	TYPE1 Xr1, Yr1;

	Tr2 = ZERO2;

	// main
	for (i = 0; i < n/(Th*N_T); i++) {
		k = i*Th*N_T;
		LOAD2_X (Xr2, &Xg[xaddr(k + iTh * N_T, ngpus, gpuid, bs)]);
		Xs2[iTh] = Xr2;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy*N_T; j < Th*N_T; j+= Ty*N_T) {
				LOAD2_A (Ar2, &Ag[(j+k+0) * lda + rowid]); 
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].x, Ar2, Tr2);
				LOAD2_A (Ar2, &Ag[(j+k+1) * lda + rowid]); 
				#if defined PREC_C || defined PREC_B || defined PREC_DB
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].z, Ar2, Tr2);
				#else
				Tr2 = MAD122 ((TYPE1&)Xs2[j/N_T].y, Ar2, Tr2);
				#endif
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th*N_T;
	if (n % N_T == 0) {
		if (k + iTh * N_T + 0 < n) {
			LOAD2_X (Xr2, &Xg[xaddr(k + iTh * N_T, ngpus, gpuid, bs)]);
			Xs2[iTh] = Xr2;
		}
	}
	else {
		if (k + iTh * N_T + 0 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 0, ngpus, gpuid, bs)]);
			(TYPE1&)Xs2[iTh].x = Xr1;
		}
		if (k + iTh * N_T + 1 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 1, ngpus, gpuid, bs)]);
		#if defined PREC_C || defined PREC_B || defined PREC_DB
			(TYPE1&)Xs2[iTh].z = Xr1;
		#else
			(TYPE1&)Xs2[iTh].y = Xr1;
		#endif
		}
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy * N_T; j < n; j += Ty*N_T) {
			if (j + 0 < n) {
				LOAD2_A (Ar2, &Ag[(j+0) * lda + rowid]); 
				Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].x, Ar2, Tr2);
			}
			if (j + 1 < n) {
				LOAD2_A (Ar2, &Ag[(j+1) * lda + rowid]); 
				#if defined PREC_C || defined PREC_B || defined PREC_DB
				Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].z, Ar2, Tr2);
				#else
				Tr2 = MAD122 ((TYPE1&)Xs2[(j-k)/N_T].y, Ar2, Tr2);
				#endif
			}
		}
	}
	__syncthreads ();

	// reduction
	Xs2[iTh] = Tr2;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) 
			Tr2 = ADD2 (Tr2, Xs2[Tx * i + iTx]);
		if (m % N_T == 0) {
			if (rowid < m) {
				LOAD2_Y (Yr2, &Yg[rowid]);
				Yr2 = MAD122 (alpha, Tr2, MUL12 (beta, Yr2));
				STORE2_Y (&Yg[rowid], Yr2);
			}
		} else {
			if (rowid+0 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+0]);
				Yr1 = MAD1 (alpha, (TYPE1&)Tr2.x, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid], Yr1);
			}
			#if defined PREC_C || defined PREC_B || defined PREC_DB
			if (rowid+1 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+1]);
				Yr1 = MAD1 (alpha, (TYPE1&)Tr2.z, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+1], Yr1);
			}
			#else
			if (rowid+1 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+1]);
				Yr1 = MAD1 (alpha, (TYPE1&)Tr2.y, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+1], Yr1);
			}
			#endif
		}
	}
#endif
}


/* ***********************************************
	kernel=v4, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_v4 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k;
	extern __shared__ TYPE4 Xs4[];
	TYPE4 Ar4, Xr4, Yr4, Tr4;
	TYPE1 Xr1, Yr1;

	Tr4 = ZERO4;

	// main
	for (i = 0; i < n/(Th*N_T); i++) {
		k = i*Th*N_T;
		LOAD4_X (Xr4, &Xg[xaddr(k + iTh * N_T, ngpus, gpuid, bs)]);
		Xs4[iTh] = Xr4;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy*N_T; j < Th*N_T; j+= Ty*N_T) {
				LOAD4_A (Ar4, &Ag[(j+k+0) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[j/N_T].x, Ar4, Tr4);
				LOAD4_A (Ar4, &Ag[(j+k+1) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[j/N_T].y, Ar4, Tr4);
				LOAD4_A (Ar4, &Ag[(j+k+2) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[j/N_T].z, Ar4, Tr4);
				LOAD4_A (Ar4, &Ag[(j+k+3) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[j/N_T].w, Ar4, Tr4);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th*N_T;
	if (n % N_T == 0) {
		if (k + iTh * N_T < n) {
			LOAD4_X (Xr4, &Xg[xaddr(k + iTh * N_T, ngpus, gpuid, bs)]);
			Xs4[iTh] = Xr4;
		}
	}
	else {
		if (k + iTh * N_T + 0 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 0, ngpus, gpuid, bs)]);
			Xs4[iTh].x = Xr1;
		}
		if (k + iTh * N_T + 1 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 1, ngpus, gpuid, bs)]);
			Xs4[iTh].y = Xr1;
		}
		if (k + iTh * N_T + 2 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 2, ngpus, gpuid, bs)]);
			Xs4[iTh].z = Xr1;
		}
		if (k + iTh * N_T + 3 < n) {
			LOAD1_X (Xr1, &Xg[xaddr(k + iTh * N_T + 3, ngpus, gpuid, bs)]);
			Xs4[iTh].w = Xr1;
		}
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy * N_T; j < n; j += Ty*N_T) {
			if (j + 0 < n) {
				LOAD4_A (Ar4, &Ag[(j+0) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[(j-k)/N_T].x, Ar4, Tr4);
			}
			if (j + 1 < n) {
				LOAD4_A (Ar4, &Ag[(j+1) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[(j-k)/N_T].y, Ar4, Tr4);
			}
			if (j + 2 < n) {
				LOAD4_A (Ar4, &Ag[(j+2) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[(j-k)/N_T].z, Ar4, Tr4);
			}
			if (j + 3 < n) {
				LOAD4_A (Ar4, &Ag[(j+3) * lda + rowid]); 
				Tr4 = MAD144 (Xs4[(j-k)/N_T].w, Ar4, Tr4);
			}
		}
	}
	__syncthreads ();

	// reduction
	Xs4[iTh] = Tr4;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) 
			Tr4 = ADD4 (Tr4, Xs4[Tx * i + iTx]);
		if (m % N_T == 0) {
			if (rowid < m) {
				LOAD4_Y (Yr4, &Yg[rowid]);
				Yr4 = MAD144 (alpha, Tr4, MUL14 (beta, Yr4));
				STORE4_Y (&Yg[rowid], Yr4);
			}
		} else {
			if (rowid+0 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+0]);
				Yr1 = MAD1 (alpha, Tr4.x, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+0], Yr1);
			}
			if (rowid+1 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+1]);
				Yr1 = MAD1 (alpha, Tr4.y, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+1], Yr1);
			}
			if (rowid+2 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+2]);
				Yr1 = MAD1 (alpha, Tr4.z, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+2], Yr1);
			}
			if (rowid+3 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+3]);
				Yr1 = MAD1 (alpha, Tr4.w, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+3], Yr1);
			}
		}
	}
#endif
}

/* ***********************************************
	kernel=s2, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_s2 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k, l;
	extern __shared__ TYPE0 Xs1[];
	TYPE1 Ar1, Xr1, Yr1, Tr1[2];

	#pragma unroll
	for (l = 0; l < N_T; l++) Tr1[l] = ZERO1;

	// main
	for (i = 0; i < n/Th; i++) {
		k = i*Th;
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < Th; j+= Ty) {
				#pragma unroll
				for (l = 0; l < N_T; l++) {
					LOAD1_A (Ar1, &Ag[(j+k) * lda + rowid+l]); 
					Tr1[l] = MAD1 (Xs1[j], Ar1, Tr1[l]);
				}
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th;
	if (k + iTh < n) {
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty) {
			#pragma unroll
			for (l = 0; l < N_T; l++) {
				LOAD1_A (Ar1, &Ag[j * lda + rowid+l]); 
				Tr1[l] = MAD1 (Xs1[j-k], Ar1, Tr1[l]);
			}
		}
	}
	__syncthreads ();

	// reduction
	#pragma unroll
	for (l = 0; l < N_T; l++) Xs1[iTh*N_T+l] = Tr1[l];
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = ADD1 (Tr1[l], Xs1[(Tx * i + iTx)*N_T+l]);
		}
		#pragma unroll
		for (l = 0; l < N_T; l++) {
			if (rowid+l < m) {
				LOAD1_Y (Yr1, &Yg[rowid+l]);
				Yr1 = MAD1 (alpha, Tr1[l], MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+l], Yr1);
			}
		}
	}
}

/* ***********************************************
	kernel=s4, trans=N
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_s4 (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k, l;
	extern __shared__ TYPE0 Xs1[];
	TYPE1 Ar1, Xr1, Yr1, Tr1[4];

	#pragma unroll
	for (l = 0; l < N_T; l++) Tr1[l] = ZERO1;

	// main
	for (i = 0; i < n/Th; i++) {
		k = i*Th;
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < Th; j+= Ty) {
				#pragma unroll
				for (l = 0; l < N_T; l++) {
					LOAD1_A (Ar1, &Ag[(j+k) * lda + rowid+l]); 
					Tr1[l] = MAD1 (Xs1[j], Ar1, Tr1[l]);
				}
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th;
	if (k + iTh < n) {
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty) {
			#pragma unroll
			for (l = 0; l < N_T; l++) {
				LOAD1_A (Ar1, &Ag[j * lda + rowid+l]); 
				Tr1[l] = MAD1 (Xs1[j-k], Ar1, Tr1[l]);
			}
		}
	}
	__syncthreads ();

	// reduction
	#pragma unroll
	for (l = 0; l < N_T; l++) Xs1[iTh*N_T+l] = Tr1[l];
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			#pragma unroll
			for (l = 0; l < N_T; l++) Tr1[l] = ADD1 (Tr1[l], Xs1[(Tx * i + iTx)*N_T+l]);
		}
		#pragma unroll
		for (l = 0; l < N_T; l++) {
			if (rowid+l < m) {
				LOAD1_Y (Yr1, &Yg[rowid+l]);
				Yr1 = MAD1 (alpha, Tr1[l], MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+l], Yr1);
			}
		}
	}
}

/* ***********************************************
	kernel=v4, trans=N for MGPU
   *********************************************** */   
__global__ void MUBLAS_GEMV_MGPU_V2_N_DEVICE_v4b (
	const SIZE_T m,
	const SIZE_T n,
	const TYPE0 alpha,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	const TYPE0 beta,
	TYPE0 *Yg,
	const uint32_t ngpus,
	const uint32_t gpuid,
	const SIZE_T bs
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE4 Ar4, Yr4, Tr4;
	TYPE1 Xr1, Yr1;

	Tr4 = ZERO4;

	// main
	for (i = 0; i < n/Th; i++) {
		k = i*Th;
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
		__syncthreads ();
		if (rowid < m) {
			for (j = iTy; j < Th; j+= Ty) {
				LOAD4_A (Ar4, &Ag[(j+k) * lda + rowid]); 
				Tr4 = MAD144 (Xs1[j], Ar4, Tr4);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = i*Th;
	if (k + iTh < n) {
		LOAD1_X (Xr1, &Xg[xaddr(k + iTh, ngpus, gpuid, bs)]);
		Xs1[iTh] = Xr1;
	}
	__syncthreads ();
	if (rowid < m) {
		for (j = k + iTy; j < n; j += Ty) {
			LOAD4_A (Ar4, &Ag[j * lda + rowid]); 
			Tr4 = MAD144 (Xs1[j-k], Ar4, Tr4);
		}
	}
	__syncthreads ();

	// reduction
	Xs1[iTh+Th*0] = Tr4.x;
	Xs1[iTh+Th*1] = Tr4.y;
	Xs1[iTh+Th*2] = Tr4.z;
	Xs1[iTh+Th*3] = Tr4.w;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr4.x = ADD1 (Tr4.x, Xs1[Tx * i + iTx + Th*0]);
			Tr4.y = ADD1 (Tr4.y, Xs1[Tx * i + iTx + Th*1]);
			Tr4.z = ADD1 (Tr4.z, Xs1[Tx * i + iTx + Th*2]);
			Tr4.w = ADD1 (Tr4.w, Xs1[Tx * i + iTx + Th*3]);
		}
		if (m % N_T == 0) {
			if (rowid < m) {
				LOAD4_Y (Yr4, &Yg[rowid]);
				Yr4 = MAD144 (alpha, Tr4, MUL14 (beta, Yr4));
				STORE4_Y (&Yg[rowid], Yr4);
			}
		} else {
			if (rowid+0 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+0]);
				Yr1 = MAD1 (alpha, Tr4.x, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+0], Yr1);
			}
			if (rowid+1 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+1]);
				Yr1 = MAD1 (alpha, Tr4.y, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+1], Yr1);
			}
			if (rowid+2 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+2]);
				Yr1 = MAD1 (alpha, Tr4.z, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+2], Yr1);
			}
			if (rowid+3 < m) {
				LOAD1_Y (Yr1, &Yg[rowid+3]);
				Yr1 = MAD1 (alpha, Tr4.w, MUL1 (beta, Yr1));
				STORE1_Y (&Yg[rowid+3], Yr1);
			}
		}
	}
#endif
}


#undef RESTRICT_A
#undef RESTRICT_X
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef LOAD1_Y
#undef LOAD2_Y
#undef LOAD4_Y
#undef STORE1_Y
#undef STORE2_Y
#undef STORE4_Y
