/* ***********************************************
 
	TRMV-LNN for CUDA 

   *********************************************** */   

#include "mublas_trmv_lnn.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_A_TRMV_LNN
#define RESTRICT_A __restrict__
#define LOAD1_A LOAD1_NC
#define LOAD2_A LOAD2_NC
#define LOAD4_A LOAD4_NC
#else
#define RESTRICT_A
#define LOAD1_A LOAD1___
#define LOAD2_A LOAD2___
#define LOAD4_A LOAD4___
#endif
#if USE_NC_X_TRMV_LNN
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif
#define STORE1_W STORE1___
#define STORE2_W STORE2___
#define STORE4_W STORE4___

/* ***********************************************
	Functions for determining Tx and Ty
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_TRMV_LNN ( mublasHandle_t* mh) {
	mh->Bx = ceilingEx ((float)mh->n / (float)(mh->N_T * mh->Tx), 1);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_TRMV_LNN ( mublasHandle_t* mh) {
	mh->ShmPerBlk = ceilingEx (2. * (float)(mh->Tx * mh->Ty * mh->N_T * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxMax_TRMV_LNN ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->n / (float)mh->N_T / (float)mh->devprop_multiProcessorCount, mh->TxStep);
}

/* ***********************************************
	Sub-functions
   *********************************************** */   
__global__ void MUBLAS_TRMV_LNN_DEVICE_final (
	const SIZE_T n,
	const TYPE0 * __restrict__ Wgl,
	const TYPE0 * __restrict__ Wgr,
	TYPE0 * Xg
) {
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx);

	if (rowid < n) 
		Xg[rowid] = ADD1 (Wgl[rowid], Wgr[rowid]);
}

/* ***********************************************
	kernel=s1, uplo=L, trans=N, diag=N
   *********************************************** */   
// Only the s1 kernel requires odd number handling (n%2!=0)

__global__ void MUBLAS_TRMV_LNN_DEVICE_s1 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx);
	SIZE_T i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE1 Ar1, Xr1, Tr1l, Tr1r;

	Tr1l = ZERO1;
	Tr1r = ZERO1;

	// main
	for (j = 0; j < (n/2)/Th; j++) {
		k = j*Th;
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
		__syncthreads ();
		for (i = iTy; i < Th; i+=Ty) {
			if (i+k < rowid+1) {
				LOAD1_A (Ar1, &Ag[(i+k) * lda + rowid]); 
				Tr1l = MAD1 (Xs1[i], Ar1, Tr1l);
			} else {
				LOAD1_A (Ar1, &Ag[(n-(i+k)) * lda + n-1-rowid]); 
				Tr1r = MAD1 (Xs1[i+Th], Ar1, Tr1r);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = j*Th;
	if (k+iTh < n/2+1+(n%2==0?0:1)) {
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
	}
	__syncthreads ();
	for (i = k+iTy; i < n/2+1; i+=Ty) {
		if (i < rowid+1) {
			if (i < n/2+(n%2?1:0)) {
				LOAD1_A (Ar1, &Ag[i * lda + rowid]); 
				Tr1l = MAD1 (Xs1[i-k], Ar1, Tr1l);
			}
		} else {
			LOAD1_A (Ar1, &Ag[(n-i) * lda + n-1-rowid]); 
			Tr1r = MAD1 (Xs1[i-k+Th], Ar1, Tr1r);
		}
	}
	__syncthreads ();

	// reduction
	Xs1[iTh] = Tr1l;
	Xs1[iTh+Th] = Tr1r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr1l = ADD1 (Tr1l, Xs1[Tx * i + iTx]);
			Tr1r = ADD1 (Tr1r, Xs1[Tx * i + iTx + Th]);
		}
		if (rowid < n) {
			STORE1_W (&Wgl[rowid], Tr1l);
			STORE1_W (&Wgr[n-1-rowid], Tr1r);
		}
	}
#endif
}

/* ***********************************************
	kernel=v2, uplo=L, trans=N, diag=N
   *********************************************** */   

__global__ void MUBLAS_TRMV_LNN_DEVICE_v2 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE2 Ar2, Tr2l, Tr2r;
	TYPE1 Xr1;

	Tr2l = ZERO2;
	Tr2r = ZERO2;

	// main
	for (j = 0; j < (n/2)/Th; j++) {
		k = j*Th;
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
		__syncthreads ();
		for (i = iTy; i < Th; i+=Ty) {
			if (i+k < min(n/2,rowid+1+1)) {
				LOAD2_A (Ar2, &Ag[(i+k) * lda + rowid]); 
				if (i+k >= rowid+1+0) (TYPE1&)Ar2.x = ZERO1;
				if (i+k >= rowid+1+1) (TYPE1&)Ar2.y = ZERO1;
				Tr2l = MAD122 (Xs1[i], Ar2, Tr2l);
			}
			if (i+k >= rowid+1) {
				LOAD2_A (Ar2, &Ag[(n-(i+k)) * lda + n-1-1-rowid]); 
				if (i+k < rowid+1+1) (TYPE1&)Ar2.x = ZERO1;
				if (i+k < rowid+1+0) (TYPE1&)Ar2.y = ZERO1;
				Tr2r = MAD122 (Xs1[i+Th], Ar2, Tr2r);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = j*Th;
	if (k+iTh < n/2+1) {
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
	}
	__syncthreads ();
	for (i = k+iTy; i < n/2+1; i+=Ty) {
		if (i < min(n/2,rowid+1+1)) {
			LOAD2_A (Ar2, &Ag[i * lda + rowid]); 
			if (i >= rowid+1+0) (TYPE1&)Ar2.x = ZERO1;
			if (i >= rowid+1+1) (TYPE1&)Ar2.y = ZERO1;
			Tr2l = MAD122 (Xs1[i-k], Ar2, Tr2l);
		}
		if (i >= rowid+1) {
			LOAD2_A (Ar2, &Ag[(n-i) * lda + n-1-1-rowid]); 
			if (i < rowid+1+1) (TYPE1&)Ar2.x = ZERO1;
			if (i < rowid+1+0) (TYPE1&)Ar2.y = ZERO1;
			Tr2r = MAD122 (Xs1[i-k+Th], Ar2, Tr2r);
		}
	}
	__syncthreads ();

	// reduction
	(TYPE2&)Xs1[iTh*N_T] = Tr2l;
	(TYPE2&)Xs1[(iTh+Th)*N_T] = Tr2r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr2l = ADD2 (Tr2l, (TYPE2&)Xs1[(Tx * i + iTx      )*N_T]);
			Tr2r = ADD2 (Tr2r, (TYPE2&)Xs1[(Tx * i + iTx + Th)*N_T]);
		}
		if (rowid < n) {
			STORE2_W (&Wgl[      rowid], Tr2l);
			STORE2_W (&Wgr[n-1-1-rowid], Tr2r);
		}
	}

#endif
}

/* ***********************************************
	kernel=v4, uplo=L, trans=N, diag=N
   *********************************************** */   

__global__ void MUBLAS_TRMV_LNN_DEVICE_v4 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_A Ag,
	const SIZE_T lda,
	const TYPE0 * RESTRICT_X Xg,
	TYPE0 *Wgl,
	TYPE0 *Wgr
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T Ty = blockDim.y;
	const SIZE_T Th = Tx * Ty;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iTy = threadIdx.y;
	const SIZE_T iTh = iTy * Tx + iTx;
	const SIZE_T rowid = (blockIdx.x * Tx + iTx) * N_T;
	SIZE_T i, j, k;
	extern __shared__ TYPE1 Xs1[];
	TYPE4 Ar4, Tr4l, Tr4r;
	TYPE1 Xr1;

	Tr4l = ZERO4;
	Tr4r = ZERO4;

	// main
	for (j = 0; j < (n/2)/Th; j++) {
		k = j*Th;
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
		__syncthreads ();
		for (i = iTy; i < Th; i+=Ty) {
			if (i+k < min(n/2,rowid+1+3)) {
				LOAD4_A (Ar4, &Ag[(i+k) * lda + rowid]); 
				if (i+k >= rowid+1+0) Ar4.x = ZERO1;
				if (i+k >= rowid+1+1) Ar4.y = ZERO1;
				if (i+k >= rowid+1+2) Ar4.z = ZERO1;
				if (i+k >= rowid+1+3) Ar4.w = ZERO1;
				Tr4l = MAD144 (Xs1[i], Ar4, Tr4l);
			}
			if (i+k >= rowid+1) {
				LOAD4_A (Ar4, &Ag[(n-(i+k)) * lda + n-1-3-rowid]); 
				if (i+k < rowid+1+3) Ar4.x = ZERO1;
				if (i+k < rowid+1+2) Ar4.y = ZERO1;
				if (i+k < rowid+1+1) Ar4.z = ZERO1;
				if (i+k < rowid+1+0) Ar4.w = ZERO1;
				Tr4r = MAD144 (Xs1[i+Th], Ar4, Tr4r);
			}
		}
		__syncthreads ();
	}

	// last one block
	k = j*Th;
	if (k+iTh < n/2+1) {
		LOAD1_X (Xr1, &Xg[k+iTh]);
		Xs1[iTh] = Xr1;
		LOAD1_X (Xr1, &Xg[n-(k+iTh)]);
		Xs1[iTh+Th] = Xr1;
	}
	__syncthreads ();
	for (i = k+iTy; i < n/2+1; i+=Ty) {
		if (i < min(n/2,rowid+1+3)) {
			LOAD4_A (Ar4, &Ag[i * lda + rowid]); 
			if (i >= rowid+1+0) Ar4.x = ZERO1;
			if (i >= rowid+1+1) Ar4.y = ZERO1;
			if (i >= rowid+1+2) Ar4.z = ZERO1;
			if (i >= rowid+1+3) Ar4.w = ZERO1;
			Tr4l = MAD144 (Xs1[i-k], Ar4, Tr4l);
		}
		if (i >= rowid+1) {
			LOAD4_A (Ar4, &Ag[(n-i) * lda + n-1-3-rowid]); 
			if (i < rowid+1+3) Ar4.x = ZERO1;
			if (i < rowid+1+2) Ar4.y = ZERO1;
			if (i < rowid+1+1) Ar4.z = ZERO1;
			if (i < rowid+1+0) Ar4.w = ZERO1;
			Tr4r = MAD144 (Xs1[i-k+Th], Ar4, Tr4r);
		}
	}
	__syncthreads ();

	// reduction
	(TYPE4&)Xs1[iTh*N_T] = Tr4l;
	(TYPE4&)Xs1[(iTh+Th)*N_T] = Tr4r;
	__syncthreads ();
	if (iTy == 0) {
		for (i = 1; i < Ty; i++) {
			Tr4l = ADD4 (Tr4l, (TYPE4&)Xs1[(Tx * i + iTx      )*N_T]);
			Tr4r = ADD4 (Tr4r, (TYPE4&)Xs1[(Tx * i + iTx + Th)*N_T]);
		}
		if (rowid < n) {
			STORE4_W (&Wgl[      rowid], Tr4l);
			STORE4_W (&Wgr[n-1-3-rowid], Tr4r);
		}
	}


#endif
}

#undef RESTRICT_A
#undef RESTRICT_X
#undef LOAD1_A
#undef LOAD2_A
#undef LOAD4_A
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef STORE1_W
#undef STORE2_W
#undef STORE4_W
